"""
Helpers.py

This module contains the OPCClient class and helper functions for interfacing with the OPC server.

The OPCClient class is built on the opcua library, which is a pure Python OPC UA stack.
It uses the OPC UA protocol (Open Platform Communications Unified Architecture)
to provide functionality for reading and writing tag values to an OPC server.

Created on May 2024.
@author: Dusan Horvath (dusan_horvath@stuba.sk)
"""
import datetime
import json
import logging
import logging.config
import webbrowser
import os

from typing import List
from urllib.parse import urlparse
from _ctypes import ArgumentError
from opcua import Client
from opcua import ua
from opcua.ua import DataValue


class OPCClient:
    """
    Class for interfacing with an OPC server using the OPC UA protocol.

    This class provides methods for reading and writing tag values to an OPC server.

    Attributes:
        _instances (dict): Holds the singleton instances of OPCClient.
        _client (Client): Instance of opcua.client.Client.
        _ip (str): IP address of the OPC server.
        _port (int): Port number of the OPC server.
        _timeout (int): Timeout for the OPC UA client in seconds.
        _opc_url (str): URL of the OPC server.
        _keep_connection (bool): Flag indicating whether to keep the connection open.
        _connected (bool): Flag indicating whether the client is connected to the OPC server.
    """
    # Hold the singleton instances of OPCClients
    _instances: dict = {}

    # Instance of opcua.client.Client
    _client: Client = None

    # IP address of the OPC server
    _ip: str
    # Port of the OPC server
    _port: int
    # Timeout in seconds for the OPC server
    _timeout: int
    # Connection string (url) for the OPC server
    _opc_url: str

    # Whether to keep the connection open
    _keep_connection: bool = False
    # Whether the connection is open
    _connected: bool = False

    def __new__(cls, server_name: str, *args, **kwargs) -> 'OPCClient':
        """
        Creates a new instance of OPCClient if it doesn't already exist for the given server_name.

        This follows the singleton pattern to ensure only one instance is created per connection string.

        Args:
            cls: Class reference
            server_name: str - OPC server name as a string

        Returns:
            OPCClient: The instance of OPCClient associated with the server_name
        """
        Logger.log_debug(f"__new__({server_name}) called.")
        if server_name not in cls._instances:
            instance = super(OPCClient, cls).__new__(cls)
            cls._instances[server_name] = instance
        return cls._instances[server_name]

    def __init__(self, server_name: str):
        """
        Initialize the OPCClient instance.

        Args:
            server_name (str): The name of the OPC server.

        Raises:
            AttributeError: If the instance has already been initialized.

        """
        # Log the initialization of the instance
        Logger.log_debug(f"__init__({server_name}) called.")

        # Ensure the instance has not been initialized before
        if not hasattr(self, '_initialized'):
            # Set the server name
            self._server_name = server_name

            # Load the OPC server parameters from the config
            self.set_config()

            # Create a new instance of the OPC Client with the actual parameters
            self.init_client()

            # Mark the instance as initialized
            self._initialized = True

            # Log warning messages about security and floating-point deviations
            Logger.log_warning("This version of OPC Client does not use any security options.")
            Logger.log_warning("This version of OPC Client has implemented but no tested floating-point deviations.")

    def name(self) -> str:
        """
        Return the name of the OPC server associated with this instance.

        Returns:
            str: The name of the OPC server.
        """
        # Return the server name
        return self._server_name

    def set_config(self):
        """
        Configure the OPC Client with actual parameters.

        This function loads the configuration for the OPC Client, validates the configuration,
        and applies the configuration to the client.

        Raises:
            ValueError: If the configuration is missing required keys.
            ValueError: If there is an error parsing the configuration.
        """
        # Load configuration
        config = Config.server(self._server_name)

        # Validate configuration
        required_keys = ['timeout', 'keepalive', 'server_url']
        if not all(key in config for key in required_keys):
            Logger.log_error("Configuration is missing required keys.")
            raise ValueError("Configuration is missing required keys.")

        try:
            # Parse configuration values
            timeout = int(config['timeout'])
            keepalive = bool(config['keepalive'])
            server_url = str(config['server_url'])
        except ValueError as e:
            Logger.log_error(f"Error parsing configuration: {e}")
            raise

        # Apply configuration to client
        self.keep_connected(keepalive)
        self.set_connection(server_url)
        self.set_timeout(timeout)

        Logger.log_info("Configuration applied successfully.")

    def set_connection(self, *args):
        """
        Set new OPC server parameters. Can accept either an IP and port or a connection string.

        Args:
            *args: Variable length argument list.

        Raises:
            ValueError: If the connection string is invalid or if the arguments are invalid.

        """
        # Check if only one argument is provided and it is a string
        if len(args) == 1 and isinstance(args[0], str):
            # Assume it's a connection string
            connection_string = args[0]
            parsed_url = urlparse(connection_string)

            # Check if the scheme is 'opc.tcp'
            if parsed_url.scheme != 'opc.tcp':
                Logger.log_warning("Invalid scheme in connection string. It must be 'opc.tcp'")
                raise ValueError("Invalid scheme in connection string. It must be 'opc.tcp'")

            # Check if the connection string has an IP address and port
            if parsed_url.hostname is None or parsed_url.port is None:
                Logger.log_warning("Invalid connection string. It must include an IP address and a port.")
                raise ValueError("Invalid connection string. It must include an IP address and a port.")

            # Set the IP, port, and OPC URL
            self._ip = parsed_url.hostname
            self._port = parsed_url.port
            self._opc_url = connection_string

        # Check if two arguments are provided and the first is a string and the second is an integer
        elif len(args) == 2 and isinstance(args[0], str) and isinstance(args[1], int):
            # Assume the first argument is IP and the second is port
            self._ip = args[0]
            self._port = args[1]
            self._opc_url = f"opc.tcp://{self._ip}:{self._port}"

        # Check if three arguments are provided and the first is a string, the second is an integer, and the third is
        # an integer
        elif len(args) == 3 and isinstance(args[0], str) and isinstance(args[1], int) and isinstance(args[2], int):
            # Set the IP, port, and timeout
            self._ip = args[0]
            self._port = args[1]
            self._timeout = args[2]

        else:
            # Raise an error if the arguments are invalid
            Logger.log_error(f"Invalid arguments for set_opc.")
            raise ValueError("Invalid arguments for set_opc.")

        # Set the OPC URL
        self._opc_url = f"opc.tcp://{self._ip}:{self._port}"

        # Log the OPC connection string
        Logger.log_info(f"OPC Connection string = {self._opc_url}")

    def keep_connected(self, keep: bool = True):
        """
        Set whether the client should keep the connection alive or reconnect every time.

        Args:
            keep (bool): If True, the client will keep the connection alive.
                         If False, the client will reconnect every time a request is made.
                         Defaults to True.

        Returns:
            None

        This function sets the `_keep_connection` attribute of the class instance.
        It also logs a message indicating the new value of the keep connection flag.
        """
        # Set the keep connection flag
        self._keep_connection = keep

        # Log the new value of the keep connection flag
        Logger.log_info(f"Keep connection is set to '{keep}' now.")

    def set_timeout(self, timeout: int):
        """
        Set the timeout for the OPC UA client in seconds.

        This function sets the timeout for the OPC UA client and performs the following steps:
        1. Closes the active connection if it exists.
        2. Initializes a new OPC UA client with the specified timeout.

        Args:
            timeout (int): The timeout value in seconds.

        Returns:
            None
        """
        # Set the timeout value
        self._timeout = timeout

        try:
            # Close the active connection if it exists
            self.disconnect()
            Logger.log_debug("Active connection closed in OPCClient.set_timeout().")
        except Exception as e:
            Logger.log_debug(f"Could not close active connection in OPCClient.set_timeout(). Error: {e}")
            pass

        finally:
            # Initialize a new OPC UA client with the specified timeout
            self.init_client()
            Logger.log_info(f"Timeout is set to '{timeout}' now.")

    def init_client(self):
        """
        Initialize the OPC UA client.

        This method initializes the OPC UA client with the provided URL and timeout.

        Args:
            self: The instance of the class.

        Returns:
            None
        """
        # Create a new instance of the OPC UA client
        self._client = Client(url=self._opc_url, timeout=self._timeout)

    @staticmethod
    def construct_full_tag_name(namespace, tags, db_name=None):
        """
        Constructs a full node name from the provided namespace, tag components, and an optional data block name.

        Args:
            namespace (str): The namespace index.
            tags (list or str): Tag or list of tags/components.
            db_name (str, optional): Optional data block name.

        Returns:
            str: The full node name.
        """
        # Join the tag components with a dot if tags is a list
        tag_path = '"."'.join(tags) if isinstance(tags, list) else tags

        # Construct the full node name based on the presence of db_name
        if db_name:
            node_name = f"ns={namespace};s=\"{db_name}\".\"{tag_path}\""
        else:
            node_name = f"ns={namespace};s=\"{tag_path}\""

        return node_name

    def read(self, tag: List[str] or str, db_name: str = '', namespace: str = '3'):
        """
        Read data from the OPC server.

        Args: tag: Union[str, List[str]] - Symbolic name of the tag to read. It can be a single string like "Tag1" or
        a list of strings like ["DB1", "Struct1", "Tag1"]. For example, if the full node name is
        "DB1"."Struct1"."Tag1", tag should be ["DB1", "Struct1", "Tag1"] or read("Tag1", db_name="Struct1",
        namespace="DB1"). db_name: str, optional - Symbolic name of the data block to read from. namespace: str,
        optional - Namespace identifier.

        Returns:
            Any - The value read from the OPC server.
        """
        self.connect()  # Connect to the OPC server
        value = None  # Initialize the value to None
        node_name = self.construct_full_tag_name(namespace, tag, db_name)  # Construct the full node name
        Logger.log_info(f"Reading from tag: {node_name}")  # Log the tag being read
        try:
            node = self._client.get_node(node_name)  # Get the node from the OPC server
            value = node.get_value()  # Retrieve the value of the node
            Logger.log_info(f"{node_name} = {value}")  # Log the value that was read
        except Exception as ex:
            Logger.log_error(f"Error during reading: {ex}")  # Log any errors that occurred during reading
        finally:
            self.disconnect(False)  # Disconnect from the OPC server
            return value

    def read_multiple(self, tag_list: List[str], default=None, namespace: str = '3') -> List:
        """
        Reads multiple tags from the PLC and returns a list of values.

        Args:
            tag_list (List[str]): List of tags to read from the PLC.
            default (Any, optional): Value to return if the read fails. Defaults to None.
            namespace (str, optional): Namespace of the tags. Defaults to '3'.

        Returns:
            List: List of values read from the PLC.

        Raises:
            Exception: If there is an error during the read operation.

        Description:
            This function connects to the PLC, iterates over each tag in the tag list,
            constructs the full tag name, gets the node, and reads the value from the node.
            If the read operation fails, the default value is returned.
            The function disconnects from the PLC and returns the list of values read.
        """
        # Connect to the PLC
        self.connect()

        # Initialize the result list
        result = []

        # Iterate over each tag in the tag list
        for tag in tag_list:
            # Construct the full tag name
            node_name = self.construct_full_tag_name(namespace, tag)

            # Get the node
            self._client.get_node(node_name)

            # Set the default value
            value = default

            try:
                # Get the node
                node = self._client.get_node(node_name)

                # Read the value from the node
                value = node.get_value()

                # Log the read value
                Logger.log_info(f"{node_name} = {value}")
            except Exception as ex:
                # Log the error if read fails
                Logger.log_error(f"Error during reading: {ex}")
            finally:
                # Append the value to the result list
                result.append(value)

        # Disconnect from the PLC
        self.disconnect(False)

        # Return the list of values read from the PLC
        return result

    def write(self, value, var_type: ua.VariantType, tag: List[str] or str, db_name: str = '',
              namespace: str = '3') -> bool:
        """
        Write a value to an OPC tag.

        Args:
            value: the value to write
            var_type: the type of the value {ua.VariantType.Int32, ua.VariantType.Float, etc.}
            tag: Symbolic name of tag to write to.
                tag may be string: "tag1" or list of strings: ["tag1", "tag2"]
                e.g., If the node name is "DB1"."Struct1"."Struct2"."Tag1",
                then tag should be ["DB1", "Struct1", "Struct2", "Tag1"]
                or
                write(... ["Struct1", "Struct2", "Tag1"], db_name="DB1")
            db_name: the name of the data block to write to (if any)
            namespace: the namespace

        Returns:
            bool: True if the write operation is successful, False otherwise.
        """
        self.connect()
        success = False
        node_name = self.construct_full_tag_name(namespace, tag, db_name)
        Logger.log_info(f"Writing {value} into tag '{node_name}'")
        try:
            node = self._client.get_node(node_name)
            variant_value = ua.Variant(value, var_type)
            node.set_value(DataValue(variant_value))
            success = True
        except Exception as ex:
            Logger.log_error(f"Exception during write '{value}' into tag '{tag}': {ex}")
        finally:
            self.disconnect(False)
            return success

    def write_multiple(self, values: List[any], var_types: List[ua.VariantType], tag_list: List[str],
                       namespaces: List[str] = None) -> List[bool]:
        """
        Write multiple values into tags.

        Args:
            values (List[any]): The list of values to write.
            var_types (List[ua.VariantType]): The list of variant types for each value.
            tag_list (List[str]): The list of tag names.
            namespaces (List[str], optional): The list of namespaces for each tag. Defaults to None.

        Returns:
            List[bool]: A list of boolean values indicating the success or failure of each write operation.

        Raises:
            ArgumentError: If the lengths of values, tag_list, and var_types are not equal.

        Description: This function writes multiple values into tags. It connects to the OPC server, checks if the
        lengths of values, tag_list, and var_types are equal, iterates over each value, tag, and namespace,
        constructs the full tag name, gets the node, creates a variant value, sets the value of the node,
        appends a boolean value to the result list indicating the success or failure of the write operation,
        logs success or failure messages, and disconnects from the OPC server before returning the result list.
        """
        # Connect to the OPC server
        self.connect()

        # Initialize the result list
        result = []

        # Check if the lengths of values, tag_list, and var_types are equal
        length = len(values)
        if length != len(tag_list) or length != len(var_types):
            Logger.log_error("Number of values, tags, and var_types must be equal!")
            self.disconnect(False)
            raise ArgumentError("Number of values, tags, and var_types must be equal!")

        # Iterate over each value, tag, and namespace
        for i in range(length):
            try:
                # Get the namespace for the current tag
                ns = '3' if ((namespaces is None) or (namespaces[i] is None)) else namespaces[i]

                # Construct the full tag name
                node_name = self.construct_full_tag_name(ns, tag_list[i])

                # Get the node
                node = self._client.get_node(node_name)

                # Create a variant value
                variant_value = ua.Variant(values[i], var_types[i])

                # Set the value of the node
                node.set_value(DataValue(variant_value))

                # Append a boolean value to the result list indicating the success of the write operation
                result.append(True)

                # Log the success message
                Logger.log_info(f"Value '{values[i]}' successfully written into tag '{tag_list[i]}'.")

            except Exception as ex:
                # Append a boolean value to the result list indicating the failure of the write operation
                result.append(False)

                # Log the exception message
                Logger.log_error(f"Exception during writing '{values[i]}' into tag '{tag_list[i]}': {ex}")

        # Disconnect from the OPC server
        self.disconnect(False)

        # Return the result list
        return result

    def connect(self):
        """
        Establishes a connection to the OPC UA server if it is not already connected.
        If the connection should be kept, the connection is not established.

        Raises:
            Exception: If the connection to the OPC UA server fails.

        Returns:
            None
        """
        # Check if the connection should be kept and if it is already connected
        if self._keep_connection and self._connected:
            # If the connection should be kept, and it is already connected, return without establishing a new
            # connection
            return

        try:
            # Set the connection status to True and establish the connection
            self._connected = True
            self._client.connect()
            # Log the successful connection
            Logger.log_info(f"Connection to OPC UA server '{self._server_name}' established.")
        except Exception as e:
            # Set the connection status to False if the connection fails
            self._connected = False
            # Log the failed connection and the error message
            Logger.log_error(f"Connection to OPC UA server '{self._server_name}' failed. : {e}")

    def disconnect(self, force: bool = True):
        """
        Close the connection to the OPC UA server.

        Args:
            force (bool, optional): If True, force close the connection regardless of keep_connection.
                                   Defaults to True.

        This function closes the connection to the OPC UA server if the connection should not be kept,
        or if the disconnection is forced. It logs a warning if the disconnection fails.

        Returns:
            None
        """
        # Check if the connection is established and either keep_connection is False or force is True
        if self._connected and (not self._keep_connection or force):
            try:
                # Disconnect from the OPC UA server
                self._client.disconnect()
                Logger.log_info("Closing connection to OPC UA server. . . (DONE)")
            except Exception as e:
                # Log a warning if the disconnection fails
                Logger.log_warning(f"Closing connection to OPC UA server. . . (FAILED): {e}")
            finally:
                # Set the connection status to False after disconnection
                self._connected = False


"""
Config class for reading configuration file
"""


class Config:
    """
    Configuration class for reading the configuration file. Contains static methods for accessing the configuration.

    Attributes:
        _configuration (dict): The configuration dictionary.
    """

    _configuration: dict = {}

    @staticmethod
    def __get_configuration() -> dict:
        """
        Reads the configuration file and returns the configuration dictionary.

        This function checks if the configuration has already been loaded.
        If not, it reads the 'config.json' file and loads the configuration.

        Returns:
            dict: The configuration dictionary.

        Raises:
            FileNotFoundError: If the 'config.json' file is not found or cannot be read.
        """
        if not Config._configuration:  # Check if configuration has already been loaded
            try:
                with open('OPC_Connector/config.json', 'r') as config_file:  # Open the config file
                    Config._configuration = json.load(config_file)  # Load the configuration
            except Exception as e:  # Handle file reading exceptions
                Logger.log_error(f"Failed to read 'OPC_Connector/config.json' file: {e}")  # Log error
                raise FileNotFoundError(
                    f"Failed to find 'OPC_Connector/config.json' file: {e}")  # Raise FileNotFoundError
        return Config._configuration  # Return the loaded configuration

    @staticmethod
    def logger() -> dict | None:
        """
        Retrieves the logger configuration from the configuration file.

        Returns:
            dict | None: The logger configuration dictionary, or None if not found.
        """
        # Get the configuration dictionary
        config_dict = Config.__get_configuration()

        # Retrieve the logger configuration from the dictionary
        logger_config = config_dict.get('logger', None)

        # If logger configuration is not found, log a warning and return None
        if logger_config is None:
            Logger.log_warning("Failed to find 'logger' configuration in 'config.json'")
            return None

        return logger_config

    @staticmethod
    def servers() -> dict | None:
        """
        Returns the server configurations from the configuration file.

        This function retrieves the server configurations dictionary from the loaded configuration.

        Returns:
            dict | None: The server configurations dictionary, or None if not found.

        Raises:
            AttributeError: If the 'opc_servers' configuration is not found in the loaded configuration.
        """
        # Get the full configuration
        cfg = Config.__get_configuration()

        # Get the server configurations from the full configuration
        servers = cfg.get('opc_servers', None)

        # Check if server configurations are found
        if servers is None:
            Logger.log_error(f"Failed to find 'opc_servers' configuration in 'config.json'")
            raise AttributeError("Failed to find 'opc_servers' configuration in 'config.json'")

        return servers

    @staticmethod
    def server(name: str) -> dict | None:
        """
        Returns the server configuration for the given 'name' from the configuration file.

        Args:
            name (str): The name of the server configuration to retrieve.

        Returns:
            dict | None: The server configuration dictionary, or None if not found.

        Raises:
            AttributeError: If the server configuration for the given name is not found.

        This function retrieves the server configuration dictionary from the loaded configuration.
        It first calls the `servers()` method to get the server configurations dictionary.
        It then uses the `get()` method to retrieve the server configuration for the given name.
        If the server configuration is not found, it logs an error and raises an AttributeError.
        Finally, it returns the server configuration dictionary.
        """
        # Get the server configurations dictionary from the loaded configuration
        cfg = Config.servers()

        # Retrieve the server configuration for the given name
        server = cfg.get(name, None)

        # If server configuration is not found, log an error and raise an AttributeError
        if server is None:
            error_message = f"Failed to find 'opc_servers.{name}' configuration in 'config.json'"
            Logger.log_error(error_message)
            raise AttributeError(error_message)

        # Return the server configuration dictionary
        return server


"""
Logger class
"""


class Logger:
    """
    Contains static methods to configure and log messages.
    """
    _logger = None

    @staticmethod
    def load():
        """
        Initialize the logger.

        This function initializes the logger by loading the logger configuration from the 'config.json' file.
        If the logger configuration is not found or is disabled, the function returns without initializing the logger.
        Otherwise, it configures the logger using the loaded configuration and sets the logger instance.

        Raises:
            KeyError: If the 'logger' configuration is not found in 'config.json'.
            Exception: If any other exception occurs during the initialization of the logger.
        """
        # Load the logger configuration from the 'config.json' file
        cfg = Config.logger()

        # If the logger configuration is not found, return without initializing the logger
        if cfg is None:
            return

        # Check if the logger is enabled
        enable = cfg.get('enable', False)

        # If the logger is disabled, return without initializing the logger
        if not enable:
            return

        try:
            # Configure the logger using the loaded configuration
            logging.config.dictConfig(cfg)

            # Set the logger instance
            Logger._logger = logging.getLogger()
        except KeyError:
            # If the 'logger' configuration is not found in 'config.json', raise a KeyError
            raise KeyError(f"Failed to find 'logger' configuration in 'config.json'")
        except Exception as e:
            # If any other exception occurs during the initialization of the logger, raise the exception
            raise e

    @staticmethod
    def log_info(message):
        """
        Logs an info message using the configured logger.

        Args:
            message (str): The message to be logged.

        This static method checks if the logger instance is initialized.
        If the logger is initialized, it logs the info message using the logger.

        Args:
            message (str): The message to be logged.
        Returns:
            None
        """
        if Logger._logger:
            Logger._logger.info(message)

    @staticmethod
    def log_error(message: str) -> None:
        """
        Log an error message using the configured logger.

        Args:
            message (str): The error message to be logged.

        This static method checks if the logger instance is initialized.
        If the logger is initialized, it logs the error message using the logger.

        Returns:
            None
        """
        # Check if the logger instance is initialized
        if Logger._logger:
            # Log the error message using the logger
            Logger._logger.error(message)

    @staticmethod
    def log_warning(message):
        """
        Log a warning message using the logger instance.

        Args:
            message (str): The warning message to be logged.
        """
        if Logger._logger:
            Logger._logger.warning(message)

    @staticmethod
    def log_debug(message):
        """
        Log a debug message.

        Args:
            message (str): The message to be logged.

        Returns:
            None
        """
        if Logger._logger:
            Logger._logger.debug(message)

    @staticmethod
    def log_critical(message):
        """
        Log a critical message using the logger instance.

        Args:
            message (str): The critical message to be logged.

        Returns:
            None
        """
        # Check if the logger instance is initialized
        if Logger._logger:
            # Log the critical message using the logger
            Logger._logger.critical(message)


"""
TicToc is a simple timer class for calculating the time difference between two
"""


class TicToc:
    """
    A utility class for timing events. It provides a simple way to measure the elapsed time
    between code execution points.

    Attributes:
        begin_time (datetime.datetime): The time when the timer was started.
    """

    begin_time: datetime.datetime = None

    @staticmethod
    def tic(b_print: bool = False):
        """
        Start the timer by recording the current time.

        Args:
            b_print (bool): If True, prints the start time to stdout.

        Returns:
            None
        """
        # Record the current time
        TicToc.begin_time = datetime.datetime.now()

        # Print the start time if b_print is True
        if b_print:
            print(f"Timer started at {TicToc.begin_time}")

    @staticmethod
    def toc(returns: bool = False, decimals: int = 3) -> float | None:
        """
        Calculates the elapsed time since `tic()` was called.

        Args:
            returns (bool): If True, returns the elapsed time as a float. If False, prints the elapsed time.
            decimals (int): The number of decimal places to round the elapsed time to.

        Returns:
            float | None: The elapsed time in seconds, rounded to the specified decimals if `returns` is True.
            None is returned if `returns` is False, and the elapsed time is printed instead.
        """
        # Check if the timer has been started
        if TicToc.begin_time is None:
            print("Timer not started. Use `tic()` to start the timer.")
            return None

        # Calculate the elapsed time
        elapsed_time = datetime.datetime.now() - TicToc.begin_time
        rounded_time = round(elapsed_time.total_seconds(), decimals)

        # Return the elapsed time or print it based on the 'returns' parameter
        if returns:
            return rounded_time
        else:
            print(f"Elapsed time: {rounded_time} seconds")


"""
Get variant type from name opcua library identified by name
"""


def get_variant_type(name: str) -> ua.VariantType:
    """
    Returns the OPC UA Variant type corresponding to the given type name.

    Args:
        name (str): The name of the OPC UA data type.

    Returns:
        ua.VariantType: The OPC UA Variant type corresponding to the given type name.

    Raises:
        ValueError: If the provided type name is not recognized.
    """
    # Mapping of type names to OPC UA Variant types
    type_mapping = {
        'bool': ua.VariantType.Boolean,
        'boolean': ua.VariantType.Boolean,
        'byte': ua.VariantType.Byte,
        'int': ua.VariantType.Int16,
        'int16': ua.VariantType.Int16,
        'int32': ua.VariantType.Int32,
        'int64': ua.VariantType.Int64,
        'float': ua.VariantType.Float,
        'real': ua.VariantType.Float,
        'double': ua.VariantType.Double,
        'string': ua.VariantType.String,
        'datetime': ua.VariantType.DateTime,
        'bytestring': ua.VariantType.ByteString,
        'guid': ua.VariantType.Guid,
        'variant': ua.VariantType.Variant,
        'datavalue': ua.VariantType.DataValue
    }

    # Convert the type name to lowercase for case-insensitive matching
    nm = name.lower()

    # Check if the type name is in the type mapping
    if nm in type_mapping:
        return type_mapping[nm]
    else:
        # Log an error and raise a ValueError if the type name is not recognized
        Logger.log_error(f"Unknown type name: {name}")
        raise ValueError(f"Unknown type name: {name}")


"""
Show documentation
"""


def open_documentation() -> None:
    """
    Opens the documentation in a web browser.

    This function uses the webbrowser module to open the documentation in a web browser.
    The documentation is located at OPC_Connector/Documentation/index.html.

    Raises:
        FileNotFoundError: If the documentation file is not found.
    """
    # Get the base path of the current file
    base_path = os.path.dirname(os.path.abspath(__file__))

    # Construct the path to the documentation file
    documentation_path = os.path.join(base_path, "Documentation", "index.html")

    # Check if the documentation file exists
    if not os.path.isfile(documentation_path):
        raise FileNotFoundError("Documentation file not found.")

    # Open the documentation in the default web browser
    webbrowser.open(documentation_path)
